// simple 2x16 LCD routines for EasPIC v8 ONLY (also handles 4x20 LCD)
// Must use 32NHz Fosc - Must use on-board LCD connections - no other pin selections are possible
// ensure that SW4-5 is off, SW4-6 is on

// This code is deliberately simple and inflexible in order to make it easy for me and new users to understand
// Amateur code - use at own risk - author accepts no responsibility for anything, anywhere, ever
// Original Code written 11 January 2021
// Author makes no claim to this original code, use any way you wish

// status 08 January 2021
//    - delays are excessive
//    - comments need improvement

// LCD connections EasyPIC v7
//
//  1  GND GND
//  2  VCC +5V
//  3   Vo contrast from 10k pot
//  4   RS RB4
//  5  R/W GND
//  6    E RB5
//  7   D0 GND
//  8   D1 GND
//  9   D2 GND
// 10   D3 GND
// 11   D4 RB0
// 12   D5 RB1
// 13   D6 RB2
// 14   D7 RB3
// 15 LED+ A      Via SW6-3 BCKL from +5V
// 16 LED- K      Via SW6-4 PWM Driven by RC2

#define _XTAL_FREQ 64000000

#include <xc.h>
#include "EP7LCD.h"

#define RS LATBbits.LATB4
#define EN LATBbits.LATB5                                                       // data clocked out on falling edge

// local function prototypes
void lcd_data4(unsigned char lcd_d);
void lcd_cmd4(char lcmd);

void lcd_clear(void){
	lcd_cmd(1);
}

// set LCD cursor position row can be 1 or 2, column can be 1 to 16
void lcd_pos(unsigned char row, unsigned char col){
  unsigned char offset;
	
	offset = 0x7f + col;                                                        // start by assuming row 1 (default)
	if(row == 2){
		offset += 0x40;                                                         // adjust to 0xc0 + col-1 for row 2
	}
	if(row == 3){
		offset += 0x14;                                                         // adjust to 0x94 + col-1 for row 3
	}
	if(row == 4){
		offset += 0x54;                                                         // adjust to 0xd4 + col-1 for row 4
	}
    lcd_cmd(offset);                                                            // first nibble
}

// set 4 bit data to LCD data lines
void lcd_data4(unsigned char lcd_d){
    LATBbits.LATB0 = lcd_d;                                                     // bit 0
    LATBbits.LATB1 = lcd_d >> 1;                                                // bit 1
    LATBbits.LATB2 = lcd_d >> 2;                                                // bit 2
    LATBbits.LATB3 = lcd_d >> 3;                                                // bit 3
}

// send 4 bit command
void lcd_cmd4(char lcmd){
	RS = 0;                                                                     // select data register
	lcd_data4(lcmd);
	EN  = 1;                                                                    // rising edge has no effect
    __delay_ms(4);                                                              // minimum 39 us - but need 4ms for all commands to complete
    EN  = 0;                                                                    // falling edge causes write
}

// send 8 bit command - requires two nibbles to be written as this is 4 bit mode
void lcd_cmd(unsigned char lcmd){
	RS = 0;                                                                     // select data register
	lcd_data4(lcmd >> 4);                                                       // high nibble
	EN  = 1;                                                                    // rising edge has no effect
    __delay_ms(4);                                                              // minimum 39 us - but need 4ms for all commands to complete
    EN  = 0;                                                                    // falling edge causes write
    __delay_ms(4);                                                              // minimum 39 us - but need 4ms for all commands to complete
	lcd_data4(lcmd);                                                            // low nibble
	EN  = 1;                                                                    // rising edge has no effect
    __delay_ms(4);                                                              // minimum 39 us - but need 4ms for all commands to complete
    EN  = 0;                                                                    // falling edge causes write
}

// write a single byte of data to LCD - requires two nibbles to be written as this is 4 bit mode
void lcd_write(unsigned char lcd_byte){

    RS = 1;                                                                     // select command register                              
    lcd_data4((lcd_byte & 0xF0) >> 4);                                          // high nibble
	EN  = 1;                                                                    // rising edge has no effect
    __delay_us(40);                                                             // minimum 37 us
    EN  = 0;                                                                    // falling edge causes write
    lcd_data4(lcd_byte & 0x0F);                                                 // low nibble
	EN  = 1;                                                                    // rising edge has no effect
    __delay_us(40);                                                             // minimum 37 us
    EN  = 0;                                                                    // falling edge causes write
}

// send null-terminated string to LCD
void lcd_text(char *ltxt){
  int i;
  
	for(i = 0; ltxt[i] != 0; i++)
	   lcd_write(ltxt[i]);
}

// initialise LCD to 4 bit mode
void lcd_init(){
    TRISB &= 0xc0;                                                              // RB4 output for RS, RB5 for EN, RB 0,1,2,3 output for data lines

    lcd_data4(0x00);
    __delay_ms(100);                                                            // minimum 40ms power-up time for LCD module
    lcd_cmd4(0x03);                                                             // function set
	__delay_ms(5);                                                              // minimum 39 us
    lcd_cmd4(0x03);                                                             // function set
	__delay_ms(11);                                                             // minimum 39 us
    lcd_cmd(0x32);                                                              // function set
    lcd_cmd(0x28);                                                              // display on
    lcd_cmd(0x0C);
    lcd_cmd(0x06);
}

