// White MIDI box PIC18F27Q43
// UART1 115200, 8, n, 1   UARTS 2, 3, 4, 5 31250 for MIDI in/out ports 1-4
// be sure to connect any unused Rx inputs to pull-up resistors
// designed for control using Sony TV IR remote control
// designed for control of MIDI keyboard, Behringer Ultra-dyne pro 9024 and Ultramizer pro 1400P

// Status 2 Feb 2021
//  - seems to be working on custom hardware and EasyPIC7

// Fixes yet to be done:
//  - ought to handle UART errors - currently errors are ignored, possibly causing occasional hang-up
//  - allow channel selections for each MIDI PORT and device?

// ideas for improvement:
//  - work out 3.5 mm IR receiver pinout - and use for this project

// Default UART pin allocation                                     changed to:
//  U1 Tx      R??    ( PORTS B and C valid)  RxyPPS = 0x20          RC6
//  U1 Rx      RC7    ( PORTS B and C valid)                        (RC7)
//  U2 Tx      R??    ( PORTS B and C valid)  RxyPPS = 0x23          RC0
//  U2 Rx      RB7    ( PORTS B and C valid)                         RC1 
//  U3 Tx      R??    ( PORTS A and B valid)  RxyPPS = 0x26          RA4
//  U3 Rx      RA7    ( PORTS A and B valid)                         RA5
//  U4 Tx      R??    ( PORTS A and B valid)  RxyPPS = 0x29          RB6
//  U4 Rx      RB5    ( PORTS B and C valid)                         RB7
//  U5 Tx      R??    ( PORTS B and C valid)  RxyPPS = 0x2c          RC4
//  U5 Rx      RA5    ( PORTS A and C valid)                         RC5

// PIC18F27Q43 pin use
//
//  1 MCLR
//  2 RA0
//  3 RA1
//  4 RA2 IR input
//  5 RA3 LCD R/W# - currently must be kept low
//  6 RA4 UART 3 TX
//  7 RA5 UART 3 RX
//  8 GND
//  9 OSC - xtal 16MHz and 22pF to gnd
// 10 OSC - xtal 16MHz and 22pF to gnd
// 11 RC0 UART 2 TX
// 12 RC1 UART 2 RX
// 13 RC2
// 14 RC3
//
// 15 RC4 UART 5 TX
// 16 RC5 UART 5 RX
// 17 RC6 UART 1 TX
// 18 RC7 UART 1 RX
// 19 GND
// 20 VCC +5V
// 21 RB0 LCD D4
// 22 RB1 LCD D5
// 23 RB2 LCD D6
// 24 RB3 LCD D7
// 25 RB4 LCD RS
// 26 RB5 LCD EN
// 27 RB6 UART 4 TX     ICSP CLK
// 28 RB7 UART 4 RX     ICSP DAT

// 2x16 character LCD connections
//
//  1  GND GND
//  2  VCC +5V
//  3   Vo contrast from 10k pot
//  4   RS RB4
//  5  R/W RA3
//  6    E RB5
//  7   D0 GND
//  8   D1 GND
//  9   D2 GND
// 10   D3 GND
// 11   D4 RB0
// 12   D5 RB1
// 13   D6 RB2
// 14   D7 RB3
// 15 LED+ A      via plug-in resistor  +5V
// 16 LED- K      GND

// MIDI connectors 5 pin DIN sockets front view
//    --- out ---                                      DIN skt
// upper    - 220R to +5V       - DIN 4 - red           ---
// middle   - gnd                                     /   2   \
// lower    - 220R from PIC Tx  - DIN 5 - yellow     / 4     5 \
//    --- in ---                                    | 1       3 |
// upper - 220R to opto pin 1   - DIN 4 - red        \    -    /
// middle - gnd                                       \  |x|  /
// lower -         opto pin 2   - DIN 5 - yellow        -----

// PICkit programming connector
// 1 Vpp   PIC pin  1 MCLR   10K to +5V
// 2 Vcc   +5V
// 3 GND   GND
// 4 PGD   PIC pin 28 RB7
// 5 PGC   PIC pin 27 RB6
// 6 AUX   nc

// CNY 17-3 Opto-Isolator
//
// 1 220R to DIN pin 4 red
// 2 DIN pin 5 yellow
// 3 nc
// 4 gnd
// 5 1k to +5 and PIC Rx
// 6 nc

// TSOP 4838 IR Receiver Module (5V supply, DC operation
//         looking front view with legs down
//   orange left   - output (30k to +) to RA2
//    black centre - Gnd
//      red right  - +5V

// casing looking from front:
//                MIDI4 Out MIDI4 In
//              --------------------------------------------
//   MIDI3 In  |                                            |
//             |                                            |
//   MIDI3 Out |                                            |
//              --------------------------------------------
//                MIDI2 In MIDI2 Out    MIDI1 In MIDI1 Out

#define _XTAL_FREQ 64000000     // 16MHz 2-pin crystal x4 PLL for 64 MHz clock

// PIC18F27Q43 Configuration Bit Settings
// CONFIG1
#pragma config FEXTOSC = HS     // External Oscillator Selection (HS (crystal oscillator) above 8 MHz)
#pragma config RSTOSC = EXTOSC_4PLL// Reset Oscillator Selection (EXTOSC with 4x PLL, with EXTOSC operating per FEXTOSC bits)
// CONFIG2
#pragma config CLKOUTEN = OFF   // Clock out Enable bit (CLKOUT function is disabled)
#pragma config PR1WAY = OFF     // PRLOCKED One-Way Set Enable bit (PRLOCKED bit can be set and cleared repeatedly)
#pragma config CSWEN = OFF      // Clock Switch Enable bit (The NOSC and NDIV bits cannot be changed by user software)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable bit (Fail-Safe Clock Monitor disabled)
// CONFIG3
#pragma config MCLRE = EXTMCLR  // MCLR Enable bit (If LVP = 0, MCLR pin is MCLR; If LVP = 1, RE3 pin function is MCLR )
#pragma config PWRTS = PWRT_64  // Power-up timer selection bits (PWRT set at 64ms)
#pragma config MVECEN = ON      // Multi-vector enable bit (Multi-vector enabled, Vector table used for interrupts)
#pragma config IVT1WAY = OFF    // IVTLOCK bit One-way set enable bit (IVTLOCKED bit can be cleared and set repeatedly)
#pragma config LPBOREN = OFF    // Low Power BOR Enable bit (Low-Power BOR disabled)
#pragma config BOREN = SBORDIS  // Brown-out Reset Enable bits (Brown-out Reset enabled , SBOREN bit is ignored)
// CONFIG4
#pragma config BORV = VBOR_1P9  // Brown-out Reset Voltage Selection bits (Brown-out Reset Voltage (VBOR) set to 1.9V)
#pragma config ZCD = OFF        // ZCD Disable bit (ZCD module is disabled. ZCD can be enabled by setting the ZCDSEN bit of ZCDCON)
#pragma config PPS1WAY = OFF    // PPSLOCK bit One-Way Set Enable bit (PPSLOCKED bit can be set and cleared repeatedly (subject to the unlock sequence))
#pragma config STVREN = ON      // Stack Full/Underflow Reset Enable bit (Stack full/underflow will cause Reset)
#pragma config LVP = OFF        // Low Voltage Programming Enable bit (HV on MCLR/VPP must be used for programming)
#pragma config XINST = OFF      // Extended Instruction Set Enable bit (Extended Instruction Set and Indexed Addressing Mode disabled)
// CONFIG5
#pragma config WDTCPS = WDTCPS_31// WDT Period selection bits (Divider ratio 1:65536; software control of WDTPS)
#pragma config WDTE = OFF       // WDT operating mode (WDT Disabled; SWDTEN is ignored)
// CONFIG6
#pragma config WDTCWS = WDTCWS_7// WDT Window Select bits (window always open (100%); software control; keyed access not required)
#pragma config WDTCCS = SC      // WDT input clock selector (Software Control)
// CONFIG7
#pragma config BBSIZE = BBSIZE_512// Boot Block Size selection bits (Boot Block size is 512 words)
#pragma config BBEN = OFF       // Boot Block enable bit (Boot block disabled)
#pragma config SAFEN = OFF      // Storage Area Flash enable bit (SAF disabled)
#pragma config DEBUG = OFF      // Background Debugger (Background Debugger disabled)
// CONFIG8
#pragma config WRTB = OFF       // Boot Block Write Protection bit (Boot Block not Write protected)
#pragma config WRTC = OFF       // Configuration Register Write Protection bit (Configuration registers not Write protected)
#pragma config WRTD = OFF       // Data EEPROM Write Protection bit (Data EEPROM not Write protected)
#pragma config WRTSAF = OFF     // SAF Write protection bit (SAF not Write Protected)
#pragma config WRTAPP = OFF     // Application Block write protection bit (Application Block not write protected)
// CONFIG10
#pragma config CP = OFF         // PFM and Data EEPROM Code Protection bit (PFM and Data EEPROM code protection disabled)

#include <xc.h>                 // always needed 
#include <stdio.h>              // allows fprint library function (among others))
#include <stdlib.h>             // not sure if this is really needed, but left just in case
#include "EP7LCD.h"             // 16x2 LCD routines compatible with EasyPIC v7

// function prototypes
void sendvol(unsigned char volume_value);
void sendprog(unsigned char program_number);
void lcd_refresh(void);
void bassboy_send(void);
void midi_off(void);
void midi_out(unsigned char midi_char);
void midi1_out(unsigned char send_char);
void midi2_out(unsigned char send_char);
void midi3_out(unsigned char send_char);
void midi4_out(unsigned char send_char);
void byte2ascii(unsigned char inbyte);
void info();
void prog_info();
void help();
void hexout(unsigned char hexval);

// Bassboy test messages - see MIDI info online for MIDI format and notes before "cmdz" function - also BassBoy manual
unsigned char midi_a[] = {0xb0, 0x3a, 0x7f                                      // Control Change 58 accent 127
                            , 0xb0, 0x35, 0x7f , 0xb0, 0x36, 0x3f               // Filter pre gain, cut-off
                            , 0xc0, 0x01                                        // select saw wave
                            , 0x90, 0x30, 0x7f , 0x90, 0x35, 0x7f               // note on 48, note on 53
                            , 0x90, 0x20, 0x7f , 0x90, 0x25, 0x7f               // note on 32, note on 35
                            , 0xc0, 0x7e                                        // select square wave
                            , 0x90, 0x30, 0x7f , 0x90, 0x35, 0x7f               // note on 48, note on 53
                            , 0x90, 0x20, 0x7f , 0x90, 0x25, 0x7f               // note on 32, note on 35
                         };                                                     // end of note array

// global variables
volatile unsigned char u1_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u1_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u1_rx_buf[257];                                                   // circular buffer for UART 1 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u2_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u2_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u2_rx_buf[257];                                                   // circular buffer for UART 2 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u3_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u3_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u3_rx_buf[257];                                                   // circular buffer for UART 3 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u4_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u4_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u4_rx_buf[257];                                                   // circular buffer for UART 4 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u5_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u5_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u5_rx_buf[257];                                                   // circular buffer for UART 5 Rx - must be 256 bytes (I added one spare)
volatile unsigned long count10ms;                                               // count tens of milliseconds
unsigned int counter;                                                           // store timer 1 value
unsigned int input_data;                                                        // assemble and store incoming bits
unsigned char bit_count;                                                        // count incoming bits
unsigned char current_state;                                                    // 1=idle 2=start 3= capture
unsigned char ir_found;                                                         // flag set in Interrupt to show message ready
unsigned char ir_enable = 1;                                                    // clear to prevent second IR read after device change
unsigned char command_id;                                                       // store for 7 bit command
unsigned char device_id;                                                        // store for 5 bit address (where 1 is TV)
unsigned char volume;                                                           // desired volume setting 0 to 24
unsigned char program;                                                          // desired program number 0 to 49 for programs 1 to 50
unsigned int timeout;                                                           // count timer time-outs for display blanks after a while
unsigned char midi_port = 5;                                                    // 0 = off, 1 to 4 = UART 2 to 5, 5 = all ports
unsigned char midi_channel = 0;                                                 // 0 to 15 for channel 1 to 16
unsigned char midi_device = 2;                                                  // 0 = MIDI in display, 1 = ultradyne, 2 = ultramizer, 3 = BassBoy synth
unsigned char maxprog = 49;                                                     // maximum number of programs for current device
unsigned char maxvol = 24;                                                      // maximum volume setting for current device
unsigned char mid_volume = 1;                                                   // mid-way volume setting for current device
unsigned int debug_uint = 0;                                                    // handy debug print option 
char conv_text[5];                                                              // temporary text buffer for conversions

// UART 1 receive interrupt
void __interrupt(irq(U1RX),high_priority) u1Int(void){
    U1RXIF = 0;                                                                 // clear interrupt condition
    u1_rx_buf[u1_rx_in_ptr] = U1RXB;                                            // put character into buffer
    u1_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 2 receive interrupt
void __interrupt(irq(U2RX),high_priority) u2Int(void){
    U2RXIF = 0;                                                                 // clear interrupt condition
    u2_rx_buf[u2_rx_in_ptr] = U2RXB;                                            // put character into buffer
    u2_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 3 receive interrupt
void __interrupt(irq(U3RX),high_priority) u3Int(void){
    U3RXIF = 0;                                                                 // clear interrupt condition
    u3_rx_buf[u3_rx_in_ptr] = U3RXB;                                            // put character into buffer
    u3_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 4 receive interrupt
void __interrupt(irq(U4RX),high_priority) u4Int(void){
    U4RXIF = 0;                                                                 // clear interrupt condition
    u4_rx_buf[u4_rx_in_ptr] = U4RXB;                                            // put character into buffer
    u4_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 5 receive interrupt
void __interrupt(irq(U5RX),high_priority) u5Int(void){
    U5RXIF = 0;                                                                 // clear interrupt condition
    u5_rx_buf[u5_rx_in_ptr] = U5RXB;                                            // put character into buffer
    u5_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// INT 0 Interrupt from IR receiver - receiver normally high, goes low on mark
void __interrupt(irq(INT0),high_priority) i0Int(void){

    switch(current_state){                                                      // state machine
      case 1:                                                                   // Idle, starting
        current_state = 2;
        INT0EDG = 1;                                                            // interrupt on rising edge
        break;
      case 2:                                                                   // start_bit, found rising edge, check length for 2.4ms
        counter = TMR1L;                                                        // also loads TMR1 H
        counter += (TMR1H * 256);                                               // combine timer 1 high and low
        counter >>= 1;
        if((counter > 2000) && (counter < 2800)){                               // acceptable uS values for 2.4ms start
            bit_count = 0;                                                      // start counting bits in
            input_data = 0;                                                     // clear old value
            current_state = 3;                                                  // start receiving bits
        }
        else{                                                                   // fault signal, reset to idle
            current_state = 1;
            INT0EDG = 0;                                                        // interrupt on falling edge
        }
        break;
      case 3:                                                                   //capture bit, start receiving data
        //check pulse length for "0" (0.6 mS space + 0.6 mS mark)
        // or "1" (0.6 mS space + 1.2 mS mark)
        counter = TMR1L;                                                        // also loads TMR1 H
        counter += (TMR1H * 256);                                               // combine timer 1 high and low
        bit_count++;
        input_data >>= 1;                                                       // make room for 0 bit

        if(counter > 3000){                                                     // acceptable uS values for zero/one discrimination
            input_data |= 0x8000;                                               // change 0 to 1
        }
        if (bit_count == 12){                                                   // complete 12 bits
            ir_found = ir_enable;                                               // indicate that a message is ready, only if IR is enabled
            input_data >>= 4;                                                   // shift 12 bits into lower bits of word
            current_state = 1;                                                  // ready to start receiving again
        }
        break;
      default:                                                                  // should never get here -
        current_state = 1;                                                      //   - but if we do - just restart
    }
    TMR1H = 0x00;                                                               // reset timer
    TMR1L = 0x00;
    INT0IF = 0;                                                                 // clear interrupt condition
}

// Timer 0 Interrupt every 10 ms
void __interrupt(irq(TMR0),high_priority) T0Int(void){
    TMR0IF = 0;                                                                 // clear interrupt
    TMR0H = 0x63;                                                               // adjust for 10ms interrupt
    TMR0L = 0xc0;                                                               // adjust for 10ms interrupt
    count10ms ++;
}

// Timer 1 Interrupt every 33 ms - clear volume/prog indication (not currently used)
void __interrupt(irq(TMR1),high_priority) T1Int(void){
    TMR1IF = 0;                                                                 // clear interrupt
        current_state = 1;                                                      // timer has overflowed - so return to idle
        timeout++;
        if(timeout > 50){                                                       // gives about 10s display timeout
            timeout = 0;                                                        // reset timeout
        }
 }

//default - catch-all interrupt handler - should never reach here
void __interrupt(irq(default),high_priority) deflt(void){
    LATA = 0xff;                                                                // light all LEDs to show something bad happened
    while(1);                                                                   // infinite loop for aiding debugging
    return;
}

// convert byte to ASCII 3 characters
void byte2ascii(unsigned char cbyte){

    conv_text[0] = (cbyte/100) + '0';
    conv_text[1] = (cbyte/10 %10) + '0';
    conv_text[2] = (cbyte %10) + '0';
    conv_text[3] = 0;                                                           // null terminate in case string wanted
}

// initialise interrupts for peripherals
void Interrupt_Init(void){
    T0CON0 = 0x90;                                                              // T0 on, 16 bit
    T0CON1 = 0x42;                                                              // fosc/4  1:4
    TMR0H = 0x63;                                                               // adjust for 10ms interrupt
    TMR0L = 0xc0;                                                               // adjust for 10ms interrupt
    TMR0IE = 1;                                                                 // enable timer 0 interrupts
    
    INT0IF = 0;                                                                 // clear INT0 interrupt
    INT0EDG = 0;                                                                // INT0 RA2 falling edge
    INT0IE = 1;                                                                 // enable INT0 on RA2 falling edge
    INT0IF = 0;                                                                 // clear INT0 interrupt

    T1CLK = 1;                                                                  // Fosc/4
    T1CON = 0x31;                                                               // timer 1, 1:8, enabled
    TMR1IF = 0;                                                                 // clear old interrupt
    TMR1H = 0x00;                                                               // maximum time-out time
    TMR1L = 0x00;
    TMR1IE = 1;                                                                 // enable timer 1 interrupt
    
}

// blocking function to send byte to serial port - required for fprintf function and similar
void putch(unsigned char send_char){                     
    
    while(! U1TXIF);                                                            // wait Tx ready
    U1TXB = send_char;                                                          // send the byte
}

// send a single MIDI byte to all ports
void midi_out(unsigned char midi_char){    
    switch(midi_port){
        case 1:
            midi1_out(midi_char);
            break;
        case 2:
            midi2_out(midi_char);
            break;
        case 3:
            midi3_out(midi_char);
            break;
        case 4:
            midi4_out(midi_char);
            break;
        case 5:
            midi1_out(midi_char);
            midi2_out(midi_char);
            midi3_out(midi_char);
            midi4_out(midi_char);
            break;
    }
    if (midi_char > 127){
        putch(13);
        putch(10);
    }
    hexout(midi_char);
    putch(' ');
}    

// send a single MIDI byte to MIDI port 1
void midi1_out(unsigned char send_char){    
    while(! U2TXIF);                                                            // wait Tx ready
    U2TXB = send_char;                                                          // send the byte
}    

// send a single MIDI byte to MIDI port 2
void midi2_out(unsigned char send_char){    
    while(! U3TXIF);                                                            // wait Tx ready
    U3TXB = send_char;                                                          // send the byte
}

// send a single MIDI byte to MIDI port 3
void midi3_out(unsigned char send_char){    
    while(! U4TXIF);                                                            // wait Tx ready
    U4TXB = send_char;                                                          // send the byte
}

// send a single MIDI byte to MIDI port 4
void midi4_out(unsigned char send_char){    
    while(! U5TXIF);                                                            // wait Tx ready
    U5TXB = send_char;                                                          // send the byte
}

// initialise desired default UARTs - could remove some Rx PPS allocations where default is used, but did not
void UART_Init(void){
    
    u1_rx_in_ptr = 0;                                                           // reset buffer pointer
    u1_rx_out_ptr = 0;                                                          // reset buffer pointer
    u2_rx_in_ptr = 0;                                                           // reset buffer pointer
    u2_rx_out_ptr = 0;                                                          // reset buffer pointer
    u3_rx_in_ptr = 0;                                                           // reset buffer pointer
    u3_rx_out_ptr = 0;                                                          // reset buffer pointer
    u4_rx_in_ptr = 0;                                                           // reset buffer pointer
    u4_rx_out_ptr = 0;                                                          // reset buffer pointer
    u5_rx_in_ptr = 0;                                                           // reset buffer pointer
    u5_rx_out_ptr = 0;                                                          // reset buffer pointer

    TRISCbits.TRISC6 = 0;                                                       // set RC6 as output for UART 1 Tx
    TRISCbits.TRISC7 = 1;                                                       // set RC7 as input for UART 1 Rx
    U1CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U1BRG = 0x8a;                                                               // desired baud rate (0x8a for 115200 baud))
    RC6PPS = 0x20;                                                              // UART 1 Tx on RC6 (default) (ports B and C valid)
    U1RXPPS = 0x17;                                                             // UART 1 Rx on RC7 (default) (ports B and C valid)
    U1CON1 = 0x80;                                                              // serial port enable
    
    TRISCbits.TRISC0 = 0;                                                       // set RC0 as output for UART 2 Tx
    TRISCbits.TRISC1 = 1;                                                       // set RC1 as input for UART 2 Rx
    U2CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U2BRG = 0x1ff;                                                              // desired baud rate (0x1ff for 31250 baud)
    RC0PPS = 0x23;                                                              // UART 2 Tx on RC0 (ports B and C valid)
    U2RXPPS = 0x11;                                                             // UART 2 Rx on RC1 (ports B and C valid)
    U2CON1 = 0x80;                                                              // serial port enable

    TRISAbits.TRISA4 = 0;                                                       // set RA4 as output for UART 3 Tx
    TRISAbits.TRISA5 = 1;                                                       // set RA5 as input for UART 3 Rx
    U3CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U3BRG = 0x1ff;                                                              // desired baud rate (0x1ff for 31250 baud)
    RA4PPS = 0x26;                                                              // UART 3 Tx on RA4 (ports A and B valid)
    U3RXPPS = 0x05;                                                             // UART 3 Rx on RA5 (ports A and B valid)
    U3CON1 = 0x80;                                                              // serial port enable
    RA7PPS = 0;                                                                 // restore old UART3 Rx default to GPIO

    TRISBbits.TRISB6 = 0;                                                       // set RB6 as output for UART 4 Tx
    TRISBbits.TRISB7 = 1;                                                       // set RB7 as input for UART 4 Rx
    U4CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U4BRG = 0x1ff;                                                              // desired baud rate (0x1ff for 31250 baud)
    RB6PPS = 0x29;                                                              // UART 4 Tx on RB6 (ports A and B valid)
    U4RXPPS = 0x0f;                                                             // UART 4 Rx on RB7 (ports B and C valid)
    U4CON1 = 0x80;                                                              // serial port enable

    TRISCbits.TRISC4 = 0;                                                       // set RC4 as output for UART 5 Tx
    TRISCbits.TRISC5 = 1;                                                       // set RC5 as input for UART 5 Rx
    U5CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U5BRG = 0x1ff;                                                              // desired baud rate (0x1ff for 31250 baud)
    RC4PPS = 0x2c;                                                              // UART 5 Tx on RC4 (ports B and C valid)
    U5RXPPS = 0x15;                                                             // UART 5 Rx on RC5 (ports A and C valid)
    U5CON1 = 0x80;                                                              // serial port enable
    RA5PPS = 0;                                                                 // restore old UART5 Rx default to GPIO
    
    __delay_ms(100);                                                            // UART settle time
    
    U1RXIE = 1;                                                                 // enable UART 1 Rx interrupt
    U2RXIE = 1;                                                                 // enable UART 2 Rx interrupt    
    U3RXIE = 1;                                                                 // enable UART 3 Rx interrupt    
    U4RXIE = 1;                                                                 // enable UART 4 Rx interrupt    
    U5RXIE = 1;                                                                 // enable UART 5 Rx interrupt    
}

// set all ports output to start with - will be modified later
void IO_Init(void){
    LATA = 0;                                                                   // all outputs low
    LATB = 0;                                                                   // all outputs low
    LATC = 0;                                                                   // all outputs low
    ANSELA = 0;                                                                 // all digital
    ANSELB = 0;                                                                 // all digital
    ANSELC = 0;                                                                 // all digital
    TRISA = 0x04;                                                               // RA2 input for IR, rest output
    INT0PPS = 0x02;                                                             // INT0 on RA2 (ports A and B valid)
}

// print byte as hex
void hexout(unsigned char hexval){
  unsigned char nibble;
  
    nibble = hexval >> 4;
    nibble += '0';
    if(nibble > '9'){
        nibble += 7;
    }
    putch(nibble);
    
    nibble = hexval & 0x0f;
    nibble += '0';
    if(nibble > '9'){
        nibble += 7;
    }
    putch(nibble);
}


// send volume 0 to 24 decimal (0 to 0x18) to MIDI
void sendvol(unsigned char volume_value){
    if(midi_device == 1){                                                       // device 1 = ultradyne
        midi_out(0xbe);                                                         // command code, channel 15
        midi_out(0x08);                                                         // command 8 - output gain
        midi_out(volume_value);                                                 // 0 to 127 decimal to output
    }
    if(midi_device == 2){                                                       // device 2 = ultramizer
        midi_out(0xbf);                                                         // command code, channel 16
        midi_out(0x4b);                                                         // command 75
        midi_out(volume_value);                                                 // 0 to 23 decimal
    }
    lcd_refresh();
}

// send program 0 to 49 decimal (0 to 0x31) to MIDI
void sendprog(unsigned char program_number){

    if(midi_device == 1){                                                       // device 1 = ultradyne
        midi_out(0xce);                                                         // program change command channel 15 (of 16)
        midi_out(program_number);                                               // program number
        __delay_ms(500);                                                        // limit update rate - DSP1400 can only update so fast
    }
    if(midi_device == 2){                                                       // device 2 = ultramizer
        midi_out(0xcf);                                                         // program change command
        midi_out(program_number);                                               // program number
        __delay_ms(500);                                                        // limit update rate - DSP1400 can only update so fast
    }
    if(midi_device == 3){                                                       // device 3 = BassBoy
        bassboy_send();
    }
    ir_found = 0;                                                               // prevent double increment
    lcd_refresh();
}

// update LCD display
void lcd_refresh(){

    lcd_pos(2,1);
    lcd_text("Port ");
    if(midi_port == 0){
        lcd_text("off ");
    }
    if(midi_port < 5){
        if(midi_port > 0){
            lcd_write(midi_port + '0');
            lcd_text("   ");
        }
    }
    if(midi_port == 5){
        lcd_text("all ");
    }

    lcd_pos(1,6);                                                               // row 1 to 2, column 1 to 16
    lcd_text("  V");
    byte2ascii(volume);
    lcd_text(conv_text);
    lcd_text(" P");
    byte2ascii(program + 1);
    lcd_text(conv_text);
    
    lcd_pos(2,10);
    lcd_text("Chan ");
    if(midi_channel < 9){                                                       // midi_channel is one less than displayed channel
        lcd_write(midi_channel + '1');
        lcd_text(" ");
    }
    else{
        lcd_text("1");
        lcd_write(midi_channel + 39);        
    }

    lcd_pos(1,1);                                                               // row 1 to 2, column 1 to 16
    lcd_text("Dev ");
    lcd_write(midi_device + '0');

    if(midi_device == 0){                                                       // device 0 output disabled
    }
    if(midi_device == 1){                                                       // device 1 = ultradyne
        maxprog = 99;                                                           // programs 1 to 100 are valid
        maxvol = 105;                                                           // volume 0 to 105 valid
        mid_volume = 51;                                                        // output level approximately equals input value on initialisation
    }
    if(midi_device == 2){                                                       // device 2 = ultramizer
        maxprog = 49;                                                           // programs 1 to 50 valid
        maxvol = 24;                                                            // volume 0 to 23 is valid
        mid_volume = 1;                                                         // would put 0, but this makes vol- button work more intuitively
    }
    if(midi_device == 3){                                                       // device 3 BassBoy
    }
    timeout = 0;                                                                // reset timeout to maximum value
}

// BassBoy MIDI send test
// insert all BassBoy links for channel 1 synth mode
// MIDI commnds include...
// 0x8x nn vv     Note off (velocity ignored)
// 0x9x nn vv     Note on
// 0xAx nn pp     Aftertouch (poly) pp=pressure
// 0xBx cc vv     Control Change vv=value
// 0xCx pp        Program Change pp=patch number
// 0xDx nn pp     Aftertouch (channel) pp=pressure
// 0xEx LL HH     Pitch Wheel llhh=pitch bend
// 0xFy ...       System Exclusive manufacturer defined
// x = MIDI channel 1 to 16
//
// 0xBx 07 vv     Volume (defsut 0x7f)
// 0xBx 35 vv     Filter pre-gain (default 0) 0=softest sound, 0x3f=hardest sound)
// 0xBx cc vv  see BassBoy manual for more details
//
// only notes 24 to 72 are recognised by BassBoy

void bassboy_send(void){
unsigned char pc;

    for(pc = 0; pc < sizeof(midi_a); pc++){                                     // step through byte by byte
        midi_out(midi_a[pc]);
        if(midi_a[pc] > 127){                                                   // control byte
            __delay_ms(1000);                                                   // note play time
        }
    }
    midi_off();                                                                 // kill all sound
}

// turn off MIDI notes
void midi_off(){
    __delay_ms(2000);                                                           // allow time to hear final note
    midi1_out(0xb0);
    midi1_out(0x7b);
    midi1_out(0x00);
}

// show useful information to terminal
void info(){
    printf("\r\nPort ");
    hexout(midi_port);
    printf(" Chan ");
    hexout(midi_channel);
    printf(" Device ");
    hexout(midi_device);
    prog_info();    
}

// show program and volume information to terminal
void prog_info(){
    printf("\r\nProgram ");
    hexout(program +1);
    printf(" Volume ");
    hexout(volume);
    printf("\r\n");    
}

// show help information to terminal
void help(){
    printf("\r\n? or h help text");
    printf("\r\nc      MIDI channel +");
    printf("\r\nd      MIDI device +");
    printf("\r\nG      device program +");
    printf("\r\ng      device program -");
    printf("\r\ni      current setting info");
    printf("\r\nm      BassBoy notes send");
    printf("\r\nn      notes send");
    printf("\r\np      MIDI port +");
    printf("\r\nr      reset PIC");
    printf("\r\nV      device volume +");
    printf("\r\nv      device volume -");
    printf("\r\n0      MIDI Ports all off");
    printf("\r\n1 to 4 MIDI Port n");
    printf("\r\n5      all MIDI Ports");
    printf("\r\n8      device mute (vol min)");
    printf("\r\n");
}

void main(void) {
  unsigned char u1_temp;
  unsigned char u2_temp;
  unsigned char u3_temp;
  unsigned char u4_temp;
  unsigned char u5_temp;

    INT0EDG = 0;                                                                // interrupt on falling edge
    current_state = 1;                                                          // Idle
    counter = 0;                                                                // initialise
    ir_found = 0;                                                               // initialise
    IO_Init();                                                                  // set up ports as desired
    Interrupt_Init();
    UART_Init();                                                                // initialise desired UART as default
    GIE = 1;                                                                    // all interrupts enabled

    printf("\r\n\nOptions:  ? c d G g h i");
    printf("\r\n       :  m n p r V v");
    printf("\r\n       :  0 1 2 3 4 5 8");
    printf("\r\n\nMIDItest 2 Feb 2021\r\n\n");

    lcd_init();                                                                 // initialise EP8 LCD
    lcd_clear();
    lcd_refresh();
    
    while(1){                                                                   // main loop
        if(u1_rx_in_ptr != u1_rx_out_ptr){                                      // is there any data in buffer?
            u1_temp = u1_rx_buf[u1_rx_out_ptr];                                 //  ... if so, then get a byte
            u1_rx_out_ptr++;                                                    // remove byte from buffer
            putch(u1_temp);                                                     // echo the byte
        }
        if(u2_rx_in_ptr != u2_rx_out_ptr){                                      // is there any data in buffer?
            u2_temp = u2_rx_buf[u2_rx_out_ptr];                                 //  ... if so, then get a byte
            u2_rx_out_ptr++;                                                    // remove byte from buffer
            if(u2_temp > 127){
                putch('\n');
                putch('\r');
            }
            putch(' ');
            putch('1');
            putch('.');
            hexout(u2_temp);                                                    // print the byte
        }
        if(u3_rx_in_ptr != u3_rx_out_ptr){                                      // is there any data in buffer?
            u3_temp = u3_rx_buf[u3_rx_out_ptr];                                 //  ... if so, then get a byte
            u3_rx_out_ptr++;                                                    // remove byte from buffer
            if(u3_temp > 127){
                putch('\n');
                putch('\r');
            }
            putch(' ');
            putch('2');
            putch('.');
            hexout(u3_temp);                                                    // print the byte
        }
        if(u4_rx_in_ptr != u4_rx_out_ptr){                                      // is there any data in buffer?
            u4_temp = u4_rx_buf[u4_rx_out_ptr];                                 //  ... if so, then get a byte
            u4_rx_out_ptr++;                                                    // remove byte from buffer
            if(u4_temp > 127){
                putch('\n');
                putch('\r');
            }
            putch(' ');
            putch('3');
            putch('.');
            hexout(u4_temp);                                                    // print the byte
        }
        if(u5_rx_in_ptr != u5_rx_out_ptr){                                      // is there any data in buffer?
            u5_temp = u5_rx_buf[u5_rx_out_ptr];                                 //  ... if so, then get a byte
            u5_rx_out_ptr++;                                                    // remove byte from buffer
            if(u5_temp > 127){
                putch('\n');
                putch('\r');
            }
            putch(' ');
            putch('4');
            putch('.');
            hexout(u5_temp);                                                    // print the byte
        }
    
        // terminal commands:   some trigger IR commands
        if (u1_temp == '?'){                                                    // help information to terminal
            u1_temp = ' ';
            help();
            info();
        }            
        if (u1_temp == '0'){                                                    // MIDI Ports off
            u1_temp = ' ';
            midi_port = 0;                                              
            lcd_refresh();
            info();
        }            
        if (u1_temp == '1'){                                                    // MIDI Port 1
            u1_temp = ' ';
            midi_port = 1;                                              
            lcd_refresh();
            info();
        }            
        if (u1_temp == '2'){                                                    // MIDI Port 2
            u1_temp = ' ';
            midi_port = 2;                                              
            lcd_refresh();
            info();
        }            
        if (u1_temp == '3'){                                                    // MIDI Port 3
            u1_temp = ' ';
            midi_port = 3;                                              
            lcd_refresh();
            info();
        }            
        if (u1_temp == '4'){                                                    // MIDI Port 4
            u1_temp = ' ';
            midi_port = 4;                                              
            lcd_refresh();
            info();
        }            
        if (u1_temp == '5'){                                                    // all MIDI Ports
            u1_temp = ' ';
            midi_port = 5;                                              
            lcd_refresh();
            info();
        }            
        if (u1_temp == '8'){                                                    // simulate IR '8' button
            u1_temp = ' ';
            input_data = 0x87;                                                  // 1-7
            ir_found = 1;
        }            
        if (u1_temp == 'c'){                                                    // MIDI channel increment
            u1_temp = ' ';
            midi_channel++;
            if(midi_channel > 15){
                midi_channel = 0;
            }
            lcd_refresh();
            info();
        }            
        if (u1_temp == 'd'){                                                    // device increment - IR '10+' button
            u1_temp = ' ';
            input_data = 0x8c;                                                  // 1-0x0c
            ir_found = 1;
        }            
        if (u1_temp == 'G'){                                                    // increment MIDI program number - IR 'P+' button
            u1_temp = ' ';
            input_data = 0x90;                                                  // 1-0x10
            ir_found = 1;
        }            
        if (u1_temp == 'g'){                                                    // decrement MIDI program number - IR 'P-' button
            u1_temp = ' ';
            input_data = 0x91;                                                  // 1-0x11
            ir_found = 1;
        }            
        if (u1_temp == 'h'){                                                    // help information to terminal
            u1_temp = ' ';
            help();
            info();
        }            
        if (u1_temp == 'i'){                                                    // information to terminal
            u1_temp = ' ';
            info();
        }            
        if (u1_temp == 'n'){                                                    // send a series of notes out to MIDI
            u1_temp = ' ';
            printf("\r\nMIDI notes send\r\n");
            midi_out(0x90 + midi_channel);                                      // note on
            midi_out(0x3c);                                                     // note number
            midi_out(0x64);                                                     // velocity (volume 64))
            __delay_ms(200);                                                    // note on time
            midi_out(0x90 + midi_channel);                                      // note on (but zero volume)
            midi_out(0x3c);                                                     // note number
            midi_out(0x00);                                                     // velocity (volume 0)
            __delay_ms(200);                                                    // note off time
            midi_out(0x90 + midi_channel);
            midi_out(0x40);
            midi_out(0x7f);
            __delay_ms(200);
            midi_out(0x90 + midi_channel);
            midi_out(0x40);
            midi_out(0x00);
            __delay_ms(200);
            midi_out(0x90 + midi_channel);
            midi_out(0x43);
            midi_out(0x7f);
            __delay_ms(200);
            midi_out(0x90 + midi_channel);
            midi_out(0x43);
            midi_out(0x00);
            __delay_ms(200);
            midi_out(0x90 + midi_channel);
            midi_out(0x47);
            midi_out(0x7f);
            __delay_ms(200);
            midi_out(0x90 + midi_channel);
            midi_out(0x47);
            midi_out(0x00);
            __delay_ms(200);
        }
        if (u1_temp == 'm'){                                                    // send a series of notes to BassBoy MIDI synth
            u1_temp = ' ';
            printf("\r\nBassBoy send MIDI 1   ....");
            bassboy_send();
            printf(" done\r\n");
        }
        if (u1_temp == 'p'){                                                    // MIDI port increment
            u1_temp = ' ';
            midi_port++;
            if(midi_port > 5){
                midi_port = 0;
            }
            lcd_refresh();
            info();
        }            
        if (u1_temp == 'r'){                                                    // reset PIC
            u1_temp = ' ';
            printf("\r\nReset   ....");
            asm("reset");
        }
        
        // handle IR messages from Sony TV remote
        if(ir_found){
            command_id = input_data & 0x7F;                                     // lower 7 bits are the command
            device_id = (input_data >> 7) & 0x0f;                               // break input data to two parts

            ir_found = 0;                                                       // clear the flag for new message
            if(device_id == 1){                                                 // device 1 is Sony TV
                switch(command_id){                                             // choose function depending upon which TV remote button pressed
                  case 7:                                                       // '8' button
                    printf(" Mute ");
                    volume = 0;                                                 // mute (nearly)
                    lcd_refresh();                                              // set LCD info according to device
                    prog_info();
                    sendvol(volume);                                            // send to MIDI
                    break;
                  case 0x0c:                                                    // 10 +
                    ir_enable = 0;                                              // prevent second switch
                    midi_device++;                                              // next MIDI device
                    if(midi_device > 3){                                        // maximum device number
                        midi_device = 0;                                        // minimum device number
                    }
                    lcd_refresh();                                              // set LCD info according to device
                    __delay_ms(100);                                            // slow the device change rate
                    ir_enable = 1;                                              // restore IR reception
                    volume = mid_volume;                                        // new device, so old volume is meaningless
                    lcd_refresh();                                              // display current volume level
                    info();
                    break;
                  case 0x10:                                                    // program +
                    if(program < maxprog){                                      // maximum permissable value
                        program++;
                        sendprog(program);                                      // send to DSP1400P
                        volume = mid_volume;                                    // new program, so old volume is meaningless
                        lcd_refresh();                                          // display current volume level
                        prog_info();
                    }
                    break;
                  case 0x11:                                                    // program -
                    if(program != 0){                                           // prevent underflow
                        program--;
                        sendprog(program);                                      // send to DSP1400P
                        volume = mid_volume;                                    // new program, so old volume is meaningless
                        lcd_refresh();                                          // display current volume level
                        prog_info();
                    }
                    break;
                  case 0x12:                                                    // volume +
                    if(volume < maxvol){                                        // maximum permissable value
                        volume++;
                        lcd_refresh();                                          // display current volume level
                        prog_info();
                        sendvol(volume);                                        // send to MIDI
                    }
                    break;
                  case 0x13:                                                    // volume -
                    if(volume != 0){                                            // prevent underflow
                        volume--;
                        lcd_refresh();                                          // display current volume level
                        prog_info();
                        sendvol(volume);                                        // send to MIDI
                    }
                    break;
                }
            }
        }        
        if (u1_temp == 'V'){                                                    // increment MIDI volume - IR 'V+' button
            u1_temp = ' ';
            input_data = 0x92;                                                  // 1-0x12
            ir_found = 1;
        }            
        if (u1_temp == 'v'){                                                    // decrement MIDI volume - IR 'V-' button
            u1_temp = ' ';
            input_data = 0x93;                                                  // 1-0x13
            ir_found = 1;
        }            

        if(debug_uint){                                                          // print out any value other than 0
            printf("\r\ndebug ");
            hexout(debug_uint >> 8);    
            hexout(debug_uint & 0x0f);    
            debug_uint = 0;                                                      // prevent repeated print
            printf("\r\n");
        }
    }
}

